#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>

#define STRMAX 1024
#define VRTMAX 2048

int dmsToDeg(char *lat, char *lon, double xy[]) {
  char dms[3][4];
  int j;

  /* check formats */
  if ((lat[0] != 'S' && lat[0] != 'N') || (lon[0] != 'W' && lon[0] != 'E'))
    return -1;

  for (j = 1; j < 7; j++)
    if (!isdigit(lat[j]))
      return -1;

  for (j = 1; j < 8; j++)
    if (!isdigit(lon[j]))
      return -1;

  /* compute longitude */
  memset(dms, '\0', sizeof(dms));
  strncpy(dms[0], &lon[1], 3);
  strncpy(dms[1], &lon[4], 2);
  strncpy(dms[2], &lon[6], 2);
  xy[0] = atof(dms[0]) + (atof(dms[1]) + atof(dms[2]) / 60.) / 60.;
  if (lon[0] == 'W')
    xy[0] = -xy[0];

  /* compute latitude */
  memset(dms, '\0', sizeof(dms));
  strncpy(dms[0], &lat[1], 2);
  strncpy(dms[1], &lat[3], 2);
  strncpy(dms[2], &lat[5], 2);
  xy[1] = atof(dms[0]) + (atof(dms[1]) + atof(dms[2]) / 60.) / 60.;
  if (lat[0] == 'S')
    xy[1] = -xy[1];

  return 0;
}

int main(int argc, char *argv[]) {
  char buf[STRMAX], w[8][128], mif[STRMAX], mid[STRMAX];
  char *key[] = {"\xca\xf0\xf3\xe3", "\xcf\xee\xeb\xe8\xe3\xee\xed"};
  double vertex[VRTMAX][2], coord[2], r, dx, dy, xmin, xmax, ymin, ymax;
  int i, k, m, id, n = -1, eof = 0;
  FILE *fp0, *fp1, *fp2;

  if (argc != 2) {
    puts("USAGE: txt2mif <txtfile>");
    exit(EXIT_FAILURE);
  }

  strcpy(mif, argv[1]);
  strcpy(strrchr(mif, '.'), ".MIF");
  strcpy(mid, argv[1]);
  strcpy(strrchr(mid, '.'), ".MID");
  if ((fp0 = fopen(argv[1], "r")) == NULL) {
    printf("can't open %s\n", argv[1]);
    exit(EXIT_FAILURE);
  }

  if ((fp1 = fopen(mif, "w")) == NULL) {
    printf("can't create %s\n", mif);
    exit(EXIT_FAILURE);
  }

  if ((fp2 = fopen(mid, "w")) == NULL) {
    printf("can't create %s\n", mid);
    exit(EXIT_FAILURE);
  }

  /* dump header */
  fprintf(fp1, "Version   300\n");
  fprintf(fp1, "Charset \"WindowsCyrillic\"\n");
  fprintf(fp1, "Delimiter \",\"\n");
  fprintf(fp1, "CoordSys Earth Projection 1, 104\n");
  fprintf(fp1, "Columns 2\n");
  fprintf(fp1, "  ID Decimal(9, 0)\n");
  fprintf(fp1, "  Type Char(10)\n");
  fprintf(fp1, "Data\n\n");

  /* process data */
  do {
    if (fgets(buf, STRMAX, fp0) == NULL) {
      m = 0;
      eof = 1;
    } else
      m = sscanf(buf, "%s %s %s %s %s %s %s %s",
		 w[0], w[1], w[2], w[3], w[4], w[5], w[6], w[7]);

    if (n >= 0) {
      if (!eof && !isdigit(buf[0]) && m == 2) {
	/* polygon vertices */
	if (dmsToDeg(w[0], w[1], vertex[n]) == 0) {
	  if (n == 0) {
	    xmin = xmax = vertex[n][0];
	    ymin = ymax = vertex[n][1];
	  } else {
	    if (xmin < vertex[n][0])
	      xmin = vertex[n][0];

	    if (xmax > vertex[n][0])
	      xmax = vertex[n][0];

	    if (ymin < vertex[n][1])
	      ymin = vertex[n][1];

	    if (ymax > vertex[n][1])
	      ymax = vertex[n][1];

	  }

	  n++;
	}

      } else {
	if (n >= 4) {
	  fprintf(fp2, "%d,\"%s\"\n", id, key[k]);
	  fprintf(fp1, "Region  1\n");
	  fprintf(fp1, "  %d\n", n);
	  for (i = 0; i < n; i++)
	    fprintf(fp1, "%f %f\n", vertex[i][0], vertex[i][1]);
	  fprintf(fp1, "    Pen (1,2,0)\n");
	  fprintf(fp1, "    Brush (1,0,16777215)\n");
	  fprintf(fp1, "    Center %f %f\n",
		  (xmin + xmax) / 2., (ymin + ymax) / 2.);
	}

	n = -1;
      }

    }

    if (!eof && isdigit(buf[0]) && m > 1) {
      id = atoi(w[0]);
      for (k = 0; k < 2; k++)
	if (strncmp(w[1], key[k], strlen(key[k])) == 0)
	  break;

      switch (k) {
      case 0:
	/* ellipse key */
	if (m == 8) {
	  w[7][8] = '\0';
	  if (dmsToDeg(w[6], w[7], coord) == 0) {
	    r = atof(w[4]);
	    dy = r / 111.195;
	    dx = dy / cos(coord[1] / 180. * M_PI);
	    fprintf(fp2, "%d,\"%s\"\n", id, key[k]);
	    fprintf(fp1, "Ellipse %f %f %f %f\n",
		    coord[0] - dx, coord[1] - dy, coord[0] + dx, coord[1] + dy);
	    fprintf(fp1, "    Pen (1,2,0)\n");
	    fprintf(fp1, "    Brush (1,0,16777215)\n");
	  }

	}
	break;

      case 1:
	/* polygon key */
	n = 0;

      }

    }

  } while (!eof);

  fclose(fp2);
  fclose(fp1);
  fclose(fp0);
  exit(EXIT_SUCCESS);
}
